/***********************************************************************************
*                               C_OUTPUT_MODULE                                    *
*                                                                                  *
*                                                                                  *
* Este  o testbench da saida dos dados que chegam aos roteadores da NOC.          *
* O nmero de roteadores que vo ser testados podem ser configurados a partir do   *
* parmetro NROT. Tambm deve-se criar novos processos para cada roteador adicio-  *
* nado. O testbench cria um arquivo de saida para cada roteador testado. Nestes    *
* arquivos aparecem os dados que foram recebidos pela porta local do roteador e    *
* informaes adicionais na seguinte sequencia:                                    *
* - 1 byte: destino do pacote                                                      *
* - 1 byte: numero de palavras                                                     *
* - 1 byte: origem do pacote                                                       *
* - 4 bytes: tempo de saida do pacote                                              *
* - N bytes: dado                                                                  *
* - 4 bytes: tempo de chegada do pacote em hexadecimal                             *
* - 1 int: tempo de chegada do pacote em inteiro                                   *
* - 1 int: tempo em que o cabealho levou para chegar no destino                   *
* - 1 int: tempo em que todo o pacote levou para chegar no destino                 *
*                                                                                  *
***********************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<errno.h>
#include <sys/timeb.h>
#include "mti.h"
#include "conv_mvl9.h"
#include "math.h"

/* nmero total de roteadores de todo o sistema */
#define NROT 25
#define TAM_FLIT 16

//interface hardware-software
typedef struct
{
	mtiSignalIdT clk, reset;
	mtiSignalIdT in_ext[NROT];
	mtiSignalIdT rx_ext[NROT];
	mtiDriverIdT ack_rx_ext[NROT];
}
signals_reg;

//declaraes globais
char g_reset, g_clk;
char g_in_ext[NROT][TAM_FLIT], g_rx_ext[NROT], g_ack_rx_ext[NROT];
int size[NROT], state[NROT];
int stamp_target[NROT][4];
char stamp_source[NROT][4][3];
unsigned long int timestamp_source[NROT], timestamp_target[NROT], timestamp_packet[NROT];
unsigned long int counter_clock;

//time global variables
struct timeb tp;
char mili[3];
char *timestamp;
long int tempo;
int segundos_inicial,milisegundos_inicial;
int segundos_final,milisegundos_final;

//deve-se ter arquivos de entrada nomeados como "inXX.txt"
//onde "XX" e o arquivo de entrada para cada uma das chaves
FILE *g_outputfile[NROT];

/* prottipos das funes */

void proc_clk_CB(signals_reg *sr);
void proc_clock_CB0(signals_reg *sr);
void proc_clock_CB1(signals_reg *sr);
void proc_clock_CB2(signals_reg *sr);
void proc_clock_CB3(signals_reg *sr);
void proc_clock_CB4(signals_reg *sr);
void proc_clock_CB5(signals_reg *sr);
void proc_clock_CB6(signals_reg *sr);
void proc_clock_CB7(signals_reg *sr);
void proc_clock_CB8(signals_reg *sr);
void proc_clock_CB9(signals_reg *sr);
void proc_clock_CB10(signals_reg *sr);
void proc_clock_CB11(signals_reg *sr);
void proc_clock_CB12(signals_reg *sr);
void proc_clock_CB13(signals_reg *sr);
void proc_clock_CB14(signals_reg *sr);
void proc_clock_CB15(signals_reg *sr);
void proc_clock_CB16(signals_reg *sr);
void proc_clock_CB17(signals_reg *sr);
void proc_clock_CB18(signals_reg *sr);
void proc_clock_CB19(signals_reg *sr);
void proc_clock_CB20(signals_reg *sr);
void proc_clock_CB21(signals_reg *sr);
void proc_clock_CB22(signals_reg *sr);
void proc_clock_CB23(signals_reg *sr);
void proc_clock_CB24(signals_reg *sr);


//prottipos das funes
void process(signals_reg *sr, int var);
void c_output_module_init(mtiRegionIdT region, char *param, mtiInterfaceListT *generics, mtiInterfaceListT *ports);
void finish(void *param);
int hex_to_dec(char value);
int byte_to_dec(char *numero);
void dec_to_hex(int numero, char *hexadecimal);
void get_timestamp_source(int var);
void get_timestamp_target(int var);
unsigned long int pow_char(char value,int position);
void calc_time(int var);
void calc_time_packet(int var);


void c_output_module_init(mtiRegionIdT region, char *param, mtiInterfaceListT *generics, mtiInterfaceListT *ports)
{
	signals_reg *sr;
	mtiProcessIdT proc_clk;
	mtiProcessIdT proc_clock[NROT];
	int i;
	char temp[15];

	sr = (signals_reg *) malloc(sizeof(signals_reg));
	//adicionando funo que  chamada nas variaes de SR
	mti_AddRestartCB(free, sr);

	//adicionando processo que  chamada ao terminar programa
	mti_AddRestartCB(finish, NULL);

//-----------------TIME--------------------------------//
	//captura o tempo
	ftime(&tp);

	//armazena o tempo inicial
	segundos_inicial=tp.time;
	milisegundos_inicial=tp.millitm;
//-----------------------------------------------------//

	//capturando portas
	sr->clk  = mti_FindPort(ports, "clk");
	sr->reset = mti_FindPort(ports, "reset");

	for(i=0;i<NROT;i++)
	{
		//capturando portas
		sprintf(temp, "in_ext%d",i);
		sr->in_ext[i] = mti_FindPort(ports, temp);
		sprintf(temp, "rx_ext%d",i);
		sr->rx_ext[i] = mti_FindPort(ports, temp);
		sprintf(temp, "ack_rx_ext%d",i);
		sr->ack_rx_ext[i] = mti_CreateDriver(mti_FindPort(ports, temp));

		//criando arquivos de saida
		sprintf(temp, "out%d.txt",i);
		if((g_outputfile[i] = fopen(temp, "w")) == NULL )
		{
			perror("fopen");
		}
	}

	//criando processo "proc_clk"
	proc_clk = mti_CreateProcess("process", (void*)proc_clk_CB, sr);

	//lista de sensitividade do processo "proc_clk"
	mti_Sensitize(proc_clk, sr->clk, MTI_EVENT);
	mti_Sensitize(proc_clk, sr->reset, MTI_EVENT);


	/* criando um processo para cada roteador conectado ao testbench */
	proc_clock[0]= mti_CreateProcess("process0", (void*)proc_clock_CB0, sr);
	proc_clock[1]= mti_CreateProcess("process1", (void*)proc_clock_CB1, sr);
	proc_clock[2]= mti_CreateProcess("process2", (void*)proc_clock_CB2, sr);
	proc_clock[3]= mti_CreateProcess("process3", (void*)proc_clock_CB3, sr);
	proc_clock[4]= mti_CreateProcess("process4", (void*)proc_clock_CB4, sr);
	proc_clock[5]= mti_CreateProcess("process5", (void*)proc_clock_CB5, sr);
	proc_clock[6]= mti_CreateProcess("process6", (void*)proc_clock_CB6, sr);
	proc_clock[7]= mti_CreateProcess("process7", (void*)proc_clock_CB7, sr);
	proc_clock[8]= mti_CreateProcess("process8", (void*)proc_clock_CB8, sr);
	proc_clock[9]= mti_CreateProcess("process9", (void*)proc_clock_CB9, sr);
	proc_clock[10]= mti_CreateProcess("process10", (void*)proc_clock_CB10, sr);
	proc_clock[11]= mti_CreateProcess("process11", (void*)proc_clock_CB11, sr);
	proc_clock[12]= mti_CreateProcess("process12", (void*)proc_clock_CB12, sr);
	proc_clock[13]= mti_CreateProcess("process13", (void*)proc_clock_CB13, sr);
	proc_clock[14]= mti_CreateProcess("process14", (void*)proc_clock_CB14, sr);
	proc_clock[15]= mti_CreateProcess("process15", (void*)proc_clock_CB15, sr);
	proc_clock[16]= mti_CreateProcess("process16", (void*)proc_clock_CB16, sr);
	proc_clock[17]= mti_CreateProcess("process17", (void*)proc_clock_CB17, sr);
	proc_clock[18]= mti_CreateProcess("process18", (void*)proc_clock_CB18, sr);
	proc_clock[19]= mti_CreateProcess("process19", (void*)proc_clock_CB19, sr);
	proc_clock[20]= mti_CreateProcess("process20", (void*)proc_clock_CB20, sr);
	proc_clock[21]= mti_CreateProcess("process21", (void*)proc_clock_CB21, sr);
	proc_clock[22]= mti_CreateProcess("process22", (void*)proc_clock_CB22, sr);
	proc_clock[23]= mti_CreateProcess("process23", (void*)proc_clock_CB23, sr);
	proc_clock[24]= mti_CreateProcess("process24", (void*)proc_clock_CB24, sr);



	// lista de sensitividade destes processos
	for(i=0;i<NROT;i++)
	{
		mti_Sensitize(proc_clock[i], sr->rx_ext[i], MTI_EVENT);
		mti_Sensitize(proc_clock[i], sr->reset, MTI_EVENT);
	}
}

// processo nico responsvel por contar o tempo global do sistema
void proc_clk_CB(signals_reg *sr)
{
	g_reset = convert_enum_to_mvl9_char(mti_GetSignalValue(sr->reset));
	g_clk   = convert_enum_to_mvl9_char(mti_GetSignalValue(sr->clk));

	if(g_reset=='0' && g_clk=='1')
		counter_clock++;
}

// cada processo  responsvel por testar os seus sinais.
// o processo  igual para todos, mudando apenas o ndice
// dos sinais que devem ser testados.
void proc_clock_CB0 (signals_reg *sr){process(sr , 0 );}
void proc_clock_CB1 (signals_reg *sr){process(sr , 1 );}
void proc_clock_CB2 (signals_reg *sr){process(sr , 2 );}
void proc_clock_CB3 (signals_reg *sr){process(sr , 3 );}
void proc_clock_CB4 (signals_reg *sr){process(sr , 4 );}
void proc_clock_CB5 (signals_reg *sr){process(sr , 5 );}
void proc_clock_CB6 (signals_reg *sr){process(sr , 6 );}
void proc_clock_CB7 (signals_reg *sr){process(sr , 7 );}
void proc_clock_CB8 (signals_reg *sr){process(sr , 8 );}
void proc_clock_CB9 (signals_reg *sr){process(sr , 9 );}
void proc_clock_CB10 (signals_reg *sr){process(sr , 10 );}
void proc_clock_CB11 (signals_reg *sr){process(sr , 11 );}
void proc_clock_CB12 (signals_reg *sr){process(sr , 12 );}
void proc_clock_CB13 (signals_reg *sr){process(sr , 13 );}
void proc_clock_CB14 (signals_reg *sr){process(sr , 14 );}
void proc_clock_CB15 (signals_reg *sr){process(sr , 15 );}
void proc_clock_CB16 (signals_reg *sr){process(sr , 16 );}
void proc_clock_CB17 (signals_reg *sr){process(sr , 17 );}
void proc_clock_CB18 (signals_reg *sr){process(sr , 18 );}
void proc_clock_CB19 (signals_reg *sr){process(sr , 19 );}
void proc_clock_CB20 (signals_reg *sr){process(sr , 20 );}
void proc_clock_CB21 (signals_reg *sr){process(sr , 21 );}
void proc_clock_CB22 (signals_reg *sr){process(sr , 22 );}
void proc_clock_CB23 (signals_reg *sr){process(sr , 23 );}
void proc_clock_CB24 (signals_reg *sr){process(sr , 24 );}



// processo que verifica se chegaram dados para uma chave
void process(signals_reg *sr, int var)
{
	int value[NROT];
	char hexa[NROT][2];

	//captura os sinais do hardware
	g_reset = convert_enum_to_mvl9_char(mti_GetSignalValue(sr->reset));
	g_clk   = convert_enum_to_mvl9_char(mti_GetSignalValue(sr->clk));
	g_rx_ext[var] = convert_enum_to_mvl9_char(mti_GetSignalValue(sr->rx_ext[var]));

	//verifica se ocorreu reset
	if(g_reset=='1')
	{
		value[var] = 0;
		convert_int_to_enums(0, &g_ack_rx_ext[var], 1);
		state[var] = 0;
	}
	else
	{
		//verifica se recebeu dado
		if(g_rx_ext[var]=='1')
		{
			//envia ack
			convert_int_to_enums(1, &g_ack_rx_ext[var], 1);

			//captura o dado recebido
			mti_GetArraySignalValue(sr->in_ext[var], g_in_ext[var]);
			value[var] = convert_enums_to_int(g_in_ext[var], TAM_FLIT);

			if(state[var]==0)
			{
				//capturando destino da mensagem
				dec_to_hex(value[var],hexa[var]);
				fprintf(g_outputfile[var], "%s ", hexa[var]);

				//capturando o tempo de chegada da mensagem no destino
				get_timestamp_target(var);

				state[var]++;
			}
			else if(state[var]==1)
			{
				//armazena o tamanho do pacote
				size[var] = value[var] - 5;

				//escrevendo nmero de flits no arquivo de saida
				dec_to_hex(value[var],hexa[var]);
				fprintf(g_outputfile[var], "%s ", hexa[var]);

				state[var]++;
			}
			else if(state[var]==2)
			{
				//escrevendo o endereco de origem da mensagem
				if(value[var]<10)
					fprintf(g_outputfile[var], "0%d ", value[var]);
				else
					fprintf(g_outputfile[var], "%d ", value[var]);

				state[var]++;
			}
			else if(state[var] >= 3 && state[var] <= 6)
			{
				if(state[var]==3) timestamp_source[var] = 0;
				
				timestamp_source[var] += value[var];

				//escrevendo timestamp da mensagem no arquivo de saida
				dec_to_hex(value[var],stamp_source[var][6 - state[var]]);
				fprintf(g_outputfile[var], "%s ", stamp_source[var][6 - state[var]]);

				//calcula tempo do pacote e escreve no respectivo arquivo de saida
				if(state[var]==6 && size[var]==0)
					calc_time(var);
				else
					state[var]++;
			}
			else
			{
				if(size[var]>0)
				{
					//escrevendo timestamp da mensagem no arquivo de saida
					dec_to_hex(value[var],hexa[var]);
					fprintf(g_outputfile[var], "%s ", hexa[var]);

					//decrementando o numero de dados restantes
					size[var]--;

					//calcula tempo do pacote e escreve no respectivo arquivo de saida
					if(state[var]>=6 && size[var]==0)
						calc_time(var);
					else
						state[var]++;
				}
				else
				{
					printf("INVALID STATE\n");
				}
			}
		}
		else
		{
			//ack permanece em zero
			convert_int_to_enums(0, &g_ack_rx_ext[var], 1);
		}
	}

	//atualiza valor da saida do ack_rx
	mti_ScheduleDriver(sr->ack_rx_ext[var], (long)g_ack_rx_ext[var], 0, MTI_INTERNAL);
}

//escreve no arquivo de saida:
//o tempo de chegada do pacote no destino em hexadecimal
//o tempo de chegada do pacote no destino em decimal
//o tempo de envio do pacote
void calc_time(int var)
{
	char palavra[3];
	char tempo[9];
	int i, time;
	unsigned long int tempo_final;

	strcpy(tempo,"");

	//proximo estado eh o inicio
	state[var] = 0;

	//escrevendo o tempo de chegada do pacote
	for(i=3;i>=0;i--)
	{
		dec_to_hex(stamp_target[var][i], palavra);
		fprintf(g_outputfile[var],"%s ",palavra);
		strcat(tempo,palavra);
	}

	//escreve o tempo de chegada do pacote em decimal
	fprintf(g_outputfile[var],"%d ", byte_to_dec(tempo));

	//calcula tempo da origem ate o destino
	time=timestamp_target[var]-timestamp_source[var];
	fprintf(g_outputfile[var],"%d ", time);

	//calcula tempo da origem ate o ultimo flit no destino
	calc_time_packet(var);
	
	//captura o tempo de simulacao em milisegundos 
	ftime(&tp);

	//armazena o tempo inicial
	segundos_final=tp.time;
	milisegundos_final=tp.millitm;

	tempo_final=(segundos_final*1000 + milisegundos_final) - (segundos_inicial*1000+milisegundos_inicial);

	fprintf(g_outputfile[var],"%ld\n",tempo_final);
	
}

//calcula o tempo da origem ate o ultimo flit no destino
void calc_time_packet(int var)
{
	int time;

	//captura o tempo de chegada do ultimo pacote
	timestamp_packet[var] = counter_clock;

	//calcula o tempo da origem ate o ultimo flit no destino
	time=timestamp_packet[var]-timestamp_source[var];

	//adicionando um ciclo para descer o sinal de recebimento
	time++;

	fprintf(g_outputfile[var],"%d ", time);
}


//processo chamado quando o simulador  fechado
//este processo simplesmente fecha os arquivos de sada
void finish(void *param)
{
	int i;
	for(i=0;i<NROT;i++)
	{
		fclose(g_outputfile[i]);
	}
}

//capturando o tempo de chegada de uma mensagem
void get_timestamp_target(int var)
{
	int i,l,j,total, temp_target[TAM_FLIT];
	unsigned long int temp_clock;

	//capturando o tempo de chegada do pacote
	timestamp_target[var] = counter_clock;

	//variavel temporaria
	temp_clock = counter_clock;

	//quebrando um long em TAMFLIT hexa
	for(i=0;i<TAM_FLIT;i++)
	{
		temp_target[i] = temp_clock % 16;
		temp_clock = temp_clock / 16;
	}

    for(i=0,j=0;i<4;i++) //4  o nmero de flits correspondente ao timestamp
    {
		total = 0;
    	for(l=0;l<(TAM_FLIT/4);l++,j++) //(TAM_FLIT/4)  o tamflit em hexa
    	{
    		total = total + temp_target[j]* pow(2,(l*4));
    	}
    	stamp_target[var][i]  = total;
    }
}

//converte um byte para decimal
int byte_to_dec(char *numero)
{
    int i, j, tam, dec, hexa,decimal=0;
    tam = strlen(numero);
    for (i = 0, j = tam - 1; i < tam; i++, j--)
    {
		hexa = hex_to_dec(numero[i]);
        dec = hexa * (pow(16, j));
        decimal = decimal + dec;
    }
    return decimal;
}

//converte qualquer valor decimal para hexadecimal
void dec_to_hex(int numero, char *hexadecimal)
{
    char hexa[20];
    int resto;

    strcpy(hexadecimal, "");
    do {
        resto = (numero % 16);
        numero = numero / 16;
        if (resto == 0)
            strcpy(hexa, "0");
        else if (resto == 1)
            strcpy(hexa, "1");
        else if (resto == 2)
            strcpy(hexa, "2");
        else if (resto == 3)
            strcpy(hexa, "3");
        else if (resto == 4)
            strcpy(hexa, "4");
        else if (resto == 5)
            strcpy(hexa, "5");
        else if (resto == 6)
            strcpy(hexa, "6");
        else if (resto == 7)
            strcpy(hexa, "7");
        else if (resto == 8)
            strcpy(hexa, "8");
        else if (resto == 9)
            strcpy(hexa, "9");
        else if (resto == 10)
            strcpy(hexa, "A");
        else if (resto == 11)
            strcpy(hexa, "B");
        else if (resto == 12)
            strcpy(hexa, "C");
        else if (resto == 13)
            strcpy(hexa, "D");
        else if (resto == 14)
            strcpy(hexa, "E");
        else if (resto == 15)
            strcpy(hexa, "F");
        strcat(hexa, hexadecimal);
        strcpy(hexadecimal, hexa);
    } while (numero != 0);

	//completando com zeros
	while(strlen(hexadecimal)<(TAM_FLIT/4))
	{
		strcpy(hexa, hexadecimal);
		strcpy(hexadecimal, "0");
		strcat(hexadecimal, hexa);
	}
}

//pow a partir de um hexadecimal
unsigned long int pow_char(char value,int position)
{
	return ( hex_to_dec(value) * pow(16,position) );
}

//hexadecimal para decimal
int hex_to_dec(char value)
{
	switch (value)
	{
		case '0':
		{ return 0; }
		case '1':
		{ return 1; }
		case '2':
		{ return 2; }
		case '3':
		{ return 3; }
		case '4':
		{ return 4; }
		case '5':
		{ return 5; }
		case '6':
		{ return 6; }
		case '7':
		{ return 7; }
		case '8':
		{ return 8; }
		case '9':
		{ return 9; }
		case 'A':
		{ return 10; }
		case 'B':
		{ return 11; }
		case 'C':
		{ return 12; }
		case 'D':
		{ return 13; }
		case 'E':
		{ return 14; }
		case 'F':
		{ return 15; }
		default:
		{
			printf("%c = ERRO NO ARQUIVO DE ENTRADA\n", value);
			return 0;
		}
	}
}
