/*----------------------------------------------------------

            Cosine Coeff. Multiplier for 2-D DCT & IDCT

  Author : CSA & VLSI Design Lab.
           Kyunghee Univ. Dept. of Electronics

  Purpose : ModelSim FLI of Random Number Generator (H.261 compliant)

  Filename : init_rand261.c
  Comment  :

------------------------------------------------------------*/

#include <stdio.h>
#include "mti.h"

/******************************************************************
 The following set of routines is used for converting STD_LOGIC
 and STD_LOGIC_VECTORs from their internal representations
 (ordinal positions in the enerated type for std_logic)
 to the character values (X, 0, 1, ...) and vice-versa.
 *******************************************************************/

char bool_char[] = { 'F', 'T'};

char convert_enum_to_bool_char(enum_bit)
int		enum_bit;
{
		if (enum_bit > 8)
			return '?';
		else
			return bool_char[ enum_bit ];
}

int convert_bool_char_to_enum(b_char)
char b_char;
{
	return((b_char=='T') ? 1 : 0);

}

char mvl9_char[] =
	{ 'U', 'X', '0', '1', 'Z', 'W', 'L', 'H', '-' };

char convert_enum_to_mvl9_char(enum_bit)
int		enum_bit;
{
		if (enum_bit > 8)
			return '?';
		else
			return mvl9_char[ enum_bit ];
}

void convert_enums_to_mvl9_string(enums, string, len)
char	*enums;
char	*string;
int		len;
{
	int i;
	for (i=0; i<len; i++) {
		string[i] = convert_enum_to_mvl9_char((int) enums[i]);
	}
	string[len] = '\0';
}

void convert_mvl9_string_to_enums(string, enums, len)
char	*string;
char	*enums;
int		len;
{
	int i;
	for (i=0; i<len; i++) {
		switch(string[i]) {
			case 'U': enums[i] = 0; break;
			case 'X': enums[i] = 1; break;
			case '0': enums[i] = 2; break;
			case '1': enums[i] = 3; break;
			case 'Z': enums[i] = 4; break;
			case 'W': enums[i] = 5; break;
			case 'L': enums[i] = 6; break;
			case 'H': enums[i] = 7; break;
			case '-': enums[i] = 8; break;
			default:
				mti_PrintMessage("error: a non-MVL9 character was found in a signal value specification.\n");
		}
	}
}

/********************************************************************
	Convert Integer to MVL9 String
*********************************************************************/
void convert_int_to_mvl9_string(num, string, len)
long	num;
char	*string;
int		len;
{
	int	i;

	for (i=0; i<len; i++)
		string[len-i-1] = ((num>>i) & 1)? '1' : '0';
	string[len] = '\0';
}

long convert_mvl9_str_to_int(string, len)
char	*string;
int		len;
{
	int		i;
	long	num=0;
	long	pow2=1;

	for (i=0; i<len; i++)
	{
		num += (string[len-i-1]=='1')? pow2 : 0;
		pow2 *= 2;
	}
	return (num);
}

long convert_int_to_enums(num, enums, len)
long	num;
char	*enums;
int		len;
{
	char buff[64];

	convert_int_to_mvl9_string(num, buff, len);
	convert_mvl9_string_to_enums(buff, enums, len);
	return(num);
}

long convert_enums_to_int(enums, len)
char	*enums;
int		len;
{
	char buff[64];
	long num;

	convert_enums_to_mvl9_string(enums, buff, len);
	num = convert_mvl9_str_to_int(buff, len);
	return (num);
}
